from functools import partial
from json import load as load_json
from os import path
from typing import Any, Dict, Optional
from yaml import FullLoader as YamlFullLoader, load as load_yaml

load_yaml = partial(load_yaml, Loader=YamlFullLoader)


class URILoader:
    LOADERS = {
        "yaml": load_yaml,
        "yml": load_yaml,
        "json": load_json,
    }

    def __init__(self):
        self.store = {}

    def load(self, uri: str) -> dict:
        if uri not in self.store:
            contents = JsonReferenceResolver.resolve(self._load_file_contents(uri), uri)
            self.store[uri] = contents

        return self.store[uri]

    def _load_file_contents(self, file_name: str) -> dict:
        file_name = path.abspath(file_name)
        file = open(file_name)
        extension = file_name.split(".")[-1]
        if extension not in URILoader.LOADERS:
            raise TypeError(f"Could not resolve uri `{file_name}`")

        return self.LOADERS[extension](file)  # type: ignore


_default_uri_loader = URILoader()


class JsonReference:
    def __init__(self, uri: str, loader: URILoader = _default_uri_loader):
        self.uri, self.reference = uri.split("#")
        self.id = uri
        self._loader = loader
        self._data: Optional[Dict] = None

    def __getitem__(self, key: str):
        return self.data[key]

    @property
    def data(self) -> dict:
        if self._data:
            return self._data

        schema = self._loader.load(self.uri)
        reference_path = self.reference.lstrip("#").strip("/").split("/")
        for item in reference_path:
            if item not in schema:
                raise KeyError(f"Could not resolve reference {self.reference}")
            schema = schema[item]

        if isinstance(schema, JsonReference):
            self._data = schema.data

        self._data = schema  # type: ignore

        return self._data

    def __repr__(self) -> str:
        return f"JsonReference({self.id})"

    def __iter__(self):
        return iter(self.data)


class JsonReferenceResolver:
    """
    Replaces $ref objects inside schema with instances of JsonReference class.
    Given the following json dict passed as `obj` parameter:
    ```
    {"a": 1, "b": {"$ref": "#/a"}}
    ```
    the `{"$ref": "#/a"}` part will be replaced by JsonReference instance, containing `{"a": 1}` value.
    """

    reference_store: Dict[str, JsonReference] = {}

    @classmethod
    def resolve(cls, obj: Any, base_uri: str = ""):
        try:
            if not isinstance(obj["$ref"], str):
                raise TypeError
            return cls._create_reference(obj, base_uri)
        except (TypeError, LookupError):
            pass

        if isinstance(obj, list):
            return [cls.resolve(item, base_uri) for item in obj]

        elif isinstance(obj, dict):
            return {key: cls.resolve(value, base_uri) for key, value in obj.items()}

        return obj

    @classmethod
    def _create_reference(cls, ref_obj: dict, base_uri: str) -> JsonReference:
        ref_str = ref_obj["$ref"]
        uri_part, ref_part = ref_str.split("#")
        if not uri_part:
            uri = base_uri
        else:
            uri = path.join(path.dirname(base_uri), uri_part)

        full_ref = uri + "#" + ref_part

        if full_ref not in cls.reference_store:
            cls.reference_store[full_ref] = JsonReference(full_ref)

        return cls.reference_store[full_ref]


class OpenApiSchema:
    def __init__(self, file_name: str, loader: URILoader = _default_uri_loader):
        self.file_name = file_name
        self._contents: Optional[dict] = None
        self.loader = loader

    @property
    def contents(self) -> dict:
        if self._contents is None:
            self._contents = self.loader.load(self.file_name)

        return self._contents

    def query(self, reference: str) -> Any:
        reference = reference.replace("\\/", "&slash;")
        reference_path = [part.replace("&slash;", "/") for part in reference.lstrip("#").strip("/").split("/")]
        schema = self.contents

        for item in reference_path:
            if item not in schema:
                raise KeyError(f"Could not resolve reference {reference}")
            schema = schema[item]

        return schema

    def __getitem__(self, item: str) -> Any:
        return self.contents[item]

    def __contains__(self, item: str) -> bool:
        return item in self.contents

    def __iter__(self):
        return iter(self.contents)


__all__ = ["JsonReference", "JsonReferenceResolver", "OpenApiSchema", "URILoader"]



from typing import Any


class ValidationError(ValueError):
    code: str
    message: str

    def __init__(self, *args, **kwargs: Any):
        if "code" in kwargs:
            self.code = kwargs["code"]

        self.context = kwargs
        self.code = self.code.format(**self.context)
        if args:
            super().__init__(*args)
        else:
            super().__init__(str(self))

    def __bool__(self) -> bool:
        return False

    def __str__(self) -> str:
        return self.message.format(**self.context)


class InvalidInputValidationError(ValidationError):
    code = "input_error"
    message = "Request body is invalid or malformed: {message}."


class TypeValidationError(ValidationError):
    code = "type_error"
    message = "Passed value must be valid {expected_type} type. " "Actual type passed was {actual_type}."


class EnumValidationError(ValidationError):
    code = "enum_error"
    message = "Passed value must be one of: {expected_values} type."


class FormatValidationError(ValidationError):
    code = "format_error"
    message = "Passed value must be valid string format: {expected_format}."


class UniqueItemsValidationError(TypeValidationError):
    code = "unique_items_error"
    message = "Passed value must contain only unique items."


class AdditionalItemsError(ValidationError):
    code = "additional_items_error"
    message = "Additional items in the array are not accepted."


class ArithmeticValidationError(ValidationError, ArithmeticError):
    code = "arithmetic_error"
    message = "Passed value is invalid."


class MultipleOfValidationError(ArithmeticValidationError):
    code = "multiple_of_error"
    message = "Passed value must be multiple of `{multiple_of}`."


class LengthValidationError(ArithmeticValidationError):
    pass


class RangeValidationError(ArithmeticValidationError):
    pass


class MinimumRangeError(RangeValidationError):
    code = "minimum_error"
    message = "Passed value must be greater or equal to set minimum `{expected_minimum}`."


class MinimumExclusiveRangeError(MinimumRangeError):
    code = "minimum_error"
    message = "Passed value must be greater than set minimum `{expected_minimum}`."


class MaximumRangeError(RangeValidationError):
    code = "maximum_error"
    message = "Passed value must be lower or equal to set maximum `{expected_maximum}`."


class MaximumExclusiveRangeError(MaximumRangeError):
    code = "maximum_exclusive_error"
    message = "Passed value must be lower than set maximum `{expected_maximum}`."


class MinimumLengthError(LengthValidationError):
    code = "minimum_error"
    message = "Passed value's length must be greater or equal to set minimum `{expected_minimum}`."


class MaximumLengthError(LengthValidationError):
    code = "maximum_error"
    message = "Passed value's length must be lower or equal to set maximum `{expected_maximum}`."


class PropertyError(ValidationError):
    code = "property_error"
    message = "Problem with property {property_name}."
    property_name = "unknown"

    def __init__(self, *args, **kwargs: Any):
        if "property_name" in kwargs:
            self.property_name = kwargs["property_name"]

        super().__init__(*args, **kwargs)


class RequiredPropertyError(PropertyError):
    code = "required_property_error"
    message = "Property `{property_name}` is required."


class PropertyValueError(PropertyError):
    code = "property_value_error:{sub_code}"
    message = "Property `{property_name}` failed to pass validation: {validation_error}."


class PropertyNameError(PropertyError):
    code = "property_name_error:{sub_code}"
    message = "Property name `{property_name}` is invalid: {validation_error}."


class AdditionalPropertyError(PropertyError):
    code = "additional_property_error"
    message = "Object does not expect additional properties. Property `{property_name}` is not allowed."


class MinimumPropertyError(PropertyError):
    code = "minimum_property_error"
    message = "The number of properties is lower than expected minimum: {expected_minimum}."


class MaximumPropertyError(PropertyError):
    code = "maximum_property_error"
    message = "The number of properties is greater than expected maximum: {expected_maximum}."


class MissingDependencyError(PropertyError):
    code = "missing_property_error"
    message = "Property `{property}` requires {dependencies} to be provided."



from decimal import Decimal
from typing import Union

from .errors import (
    MaximumExclusiveRangeError,
    MaximumRangeError,
    MinimumExclusiveRangeError,
    MinimumRangeError,
    MultipleOfValidationError,
)

Number = Union[int, float, Decimal]


def validate_multiple_of(value: Number, multiple_of: Number) -> Number:
    if not value % multiple_of == 0:  # type: ignore
        raise MultipleOfValidationError(multiple_of=multiple_of)

    return value


def validate_minimum(value: Number, expected_minimum: Number) -> Number:
    if value >= expected_minimum:
        return value

    raise MinimumRangeError(expected_minimum=expected_minimum)


def validate_exclusive_minimum(value: Number, expected_minimum: Number) -> Number:
    if value > expected_minimum:
        return value

    raise MinimumExclusiveRangeError(expected_minimum=expected_minimum)


def validate_maximum(value: Number, expected_maximum: Number) -> Number:
    if value <= expected_maximum:
        return value

    raise MaximumRangeError(expected_maximum=expected_maximum)


def validate_exclusive_maximum(value: Number, expected_maximum: Number) -> Number:
    if value < expected_maximum:
        return value

    raise MaximumExclusiveRangeError(expected_maximum=expected_maximum)


__all__ = [
    "validate_exclusive_maximum",
    "validate_exclusive_minimum",
    "validate_maximum",
    "validate_minimum",
    "validate_multiple_of",
]



from functools import partial
from typing import Any, Callable, Dict, List

from .array_validators import (
    validate_items,
    validate_maximum_items,
    validate_minimum_items,
    validate_tuple,
    validate_unique_items,
)
from .combining_validators import validate_all_of, validate_any_of, validate_not, validate_one_of
from .number_validators import (
    validate_exclusive_maximum,
    validate_exclusive_minimum,
    validate_maximum,
    validate_minimum,
    validate_multiple_of,
)
from .object_validators import (
    validate_object_dependencies,
    validate_object_maximum_properties,
    validate_object_minimum_properties,
    validate_object_properties,
    validate_object_property_names,
    validate_object_required_properties,
)
from .string_validators import (
    validate_maximum_length,
    validate_minimum_length,
    validate_string_format,
    validate_string_pattern,
)
from .type_validators import (
    validate_array,
    validate_boolean,
    validate_enum,
    validate_integer,
    validate_number,
    validate_object,
    validate_string,
)


def build_validator_from_schema(schema: Dict[str, Any]) -> Callable:
    if "anyOf" in schema:
        validators = [build_validator_from_schema(item) for item in schema["anyOf"]]
        return partial(validate_any_of, validators=validators)

    if "oneOf" in schema:
        validators = [build_validator_from_schema(item) for item in schema["oneOf"]]
        return partial(validate_one_of, validators=validators)

    if "allOf" in schema:
        validators = [build_validator_from_schema(item) for item in schema["allOf"]]
        return partial(validate_all_of, validators=validators)

    if "not" in schema:
        return partial(validate_not, validator=build_validator_from_schema(schema["not"]))

    if "type" in schema:
        return _build_validator_for_type(schema["type"], schema)

    if "enum" in schema:
        return _build_enum_validator(schema)

    raise ValueError("Could not build validator for passed schema")


def _build_validator_for_type(schema_type: str, definition: Dict[str, Any]) -> Callable:
    if schema_type == "boolean":
        return _build_boolean_validator()

    if schema_type == "integer":
        return _build_number_validator(validate_integer, definition)

    if schema_type == "number":
        return _build_number_validator(validate_number, definition)

    if schema_type == "string":
        return _build_string_validator(definition)

    if schema_type == "array":
        return _build_array_validator(definition)

    if schema_type == "object":
        return _build_object_validator(definition)

    # default to string validation
    return _build_string_validator(definition)


def _build_string_validator(definition: Dict[str, Any]) -> Callable:
    validators = [validate_string]
    if "enum" in definition:
        validators.append(_build_enum_validator(definition))

    if "format" in definition:
        validators.append(partial(validate_string_format, format_name=definition["format"]))

    if "pattern" in definition:
        validators.append(partial(validate_string_pattern, pattern=definition["pattern"]))

    if "minLength" in definition:
        validators.append(partial(validate_minimum_length, expected_minimum=definition["minLength"]))

    if "maxLength" in definition:
        validators.append(partial(validate_maximum_length, expected_maximum=definition["maxLength"]))

    return partial(validate_all_of, validators=validators)


def _build_enum_validator(definition: Dict[str, Any]) -> Callable:
    return partial(validate_enum, values=definition["enum"])


def _build_boolean_validator() -> Callable:
    return validate_boolean


def _build_number_validator(base_validator: Callable, definition: Dict[str, Any]) -> Callable:
    validators = [base_validator]
    _add_range_validators(validators, definition)
    if "multipleOf" in definition:
        validators.append(partial(validate_multiple_of, multiple_of=definition["multipleOf"]))

    if "enum" in definition:
        validators.append(_build_enum_validator(definition))

    return partial(validate_all_of, validators=validators)


def _build_array_validator(definition: Dict[str, Any]) -> Callable:
    validators = [validate_array]
    if "minItems" in definition:
        validators.append(partial(validate_minimum_items, expected_minimum=definition["minItems"]))
    if "maxItems" in definition:
        validators.append(partial(validate_maximum_items, expected_maximum=definition["maxItems"]))

    if "uniqueItems" in definition and definition["uniqueItems"]:
        validators.append(validate_unique_items)

    if "items" in definition:
        if isinstance(definition["items"], list):
            _build_tuple_validator(definition, validators)
        elif isinstance(definition["items"], dict):
            validators.append(partial(validate_items, item_validator=build_validator_from_schema(definition["items"]),))

    return partial(validate_all_of, validators=validators)


def _build_tuple_validator(definition: Dict[str, Any], validators: List[Callable]) -> None:
    items_validators = [build_validator_from_schema(item_schema) for item_schema in definition["items"]]
    if "additionalItems" in definition:
        if isinstance(definition["additionalItems"], bool):
            validators.append(
                partial(
                    validate_tuple, item_validator=items_validators, additional_items=definition["additionalItems"],
                )
            )
        else:
            validators.append(
                partial(
                    validate_tuple,
                    item_validator=items_validators,
                    additional_items=build_validator_from_schema(definition["additionalItems"]),
                )
            )
    else:
        validators.append(partial(validate_tuple, item_validator=items_validators, additional_items=True))


def _build_object_validator(definition: Dict[str, Any]) -> Callable:
    validators = [validate_object]

    if "propertyNames" in definition:
        definition["propertyNames"]["type"] = "string"
        validators.append(
            partial(
                validate_object_property_names, property_names=build_validator_from_schema(definition["propertyNames"]),
            )
        )

    if "minProperties" in definition:
        validators.append(partial(validate_object_minimum_properties, expected_minimum=definition["minProperties"],))

    if "maxProperties" in definition:
        validators.append(partial(validate_object_maximum_properties, expected_maximum=definition["maxProperties"],))

    if "required" in definition:
        validators.append(partial(validate_object_required_properties, required_properties=definition["required"],))

    if "dependencies" in definition:
        validators.append(partial(validate_object_dependencies, dependencies=definition["dependencies"]))

    property_validator_settings: Dict[str, Any] = {}

    if "properties" in definition:
        properties_validators = {
            property_name: build_validator_from_schema(property_schema)
            for property_name, property_schema in definition["properties"].items()
        }

        property_validator_settings["properties"] = properties_validators

    if "additionalProperties" in definition:
        if isinstance(definition["additionalProperties"], bool):
            property_validator_settings["additional_properties"] = definition["additionalProperties"]
        else:
            property_validator_settings["additional_properties"] = build_validator_from_schema(
                definition["additionalProperties"]
            )

    if "patternProperties" in definition:
        property_validator_settings["pattern_properties"] = {
            key: build_validator_from_schema(value) for key, value in definition["patternProperties"].items()
        }

    if property_validator_settings:
        validators.append(partial(validate_object_properties, **property_validator_settings))

    return partial(validate_all_of, validators=validators)


def _add_range_validators(validators: List[Callable], definition: Dict[str, Any]) -> None:
    if "minimum" in definition:
        validators.append(partial(validate_minimum, expected_minimum=definition["minimum"]))
    if "inclusiveMinimum" in definition:
        validators.append(partial(validate_minimum, expected_minimum=definition["inclusiveMinimum"]))

    if "maximum" in definition:
        validators.append(partial(validate_maximum, expected_maximum=definition["maximum"]))
    if "inclusiveMaximum" in definition:
        validators.append(partial(validate_maximum, expected_maximum=definition["inclusiveMaximum"]))

    if "exclusiveMinimum" in definition:
        validators.append(partial(validate_exclusive_minimum, expected_minimum=definition["exclusiveMinimum"],))
    if "exclusiveMaximum" in definition:
        validators.append(partial(validate_exclusive_maximum, expected_maximum=definition["exclusiveMaximum"],))



import inspect
from typing import Any

from chocs.http_message import CompositeHttpMessage
from chocs.http_request import HttpRequest
from chocs.http_response import HttpResponse
from chocs.middleware import Middleware, MiddlewareHandler
from chocs.routing import Route


class ParsedBodyMiddleware(Middleware):
    def __init__(self, strict: bool = True):
        self.strict = strict

    def handle(self, request: HttpRequest, next: MiddlewareHandler) -> HttpResponse:
        route = request.route
        assert isinstance(route, Route)
        if "parsed_body" in route.attributes:
            self._map_parsed_body(request, route)

        return next(request)

    def _map_parsed_body(self, request: HttpRequest, route: Route) -> None:
        if not inspect.isclass(route.attributes["parsed_body"]):
            return

        if not isinstance(request.parsed_body, CompositeHttpMessage):
            return

        body = request.parsed_body

        strict = route.attributes["strict"] if "strict" in route.attributes else self.strict
        constructor = route.attributes["parsed_body"]
        request._parsed_body = None

        if not strict:

            def _get_non_strict_parsed_body() -> Any:

                instance = constructor.__new__(constructor)
                for prop_name, prop_value in body.items():
                    setattr(instance, prop_name, prop_value)

                if hasattr(instance, "__post_init__"):
                    instance.__post_init__()

                return instance

            request._parsed_body_getter = _get_non_strict_parsed_body

            return

        def _get_strict_parsed_body() -> Any:
            return constructor(**body)

        request._parsed_body_getter = _get_strict_parsed_body


__all__ = ["ParsedBodyMiddleware"]



import base64
from cgi import parse_header
from io import BytesIO
from typing import Any, Dict
from urllib.parse import quote_plus

from chocs.http_headers import HttpHeaders
from chocs.http_query_string import HttpQueryString
from chocs.http_request import HttpRequest
from chocs.http_response import HttpResponse
from chocs.http_status import HttpStatus

from .serverless import ServerlessFunction

TEXT_MIME_TYPES = [
    "application/json",
    "application/javascript",
    "application/xml",
    "application/vnd.api+json",
    "image/svg+xml",
]

AwsEvent = Dict[str, Any]
AwsContext = Dict[str, Any]


class AwsServerlessFunction(ServerlessFunction):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.middleware_enabled = True

    def __call__(self, *args):
        event: AwsEvent = args[0]
        context: AwsContext = args[0]

        if event.get("source") in [
            "aws.events",
            "serverless-plugin-warmup",
        ]:  # lambda warmup should be ignored
            return {
                "statusCode": int(HttpStatus.CONTINUE),
            }
        request = create_http_request_from_aws_event(event, context)
        return format_response_to_aws(event, super().__call__(request))


def is_http_api_lambda(event: AwsEvent) -> bool:
    if event.get("version") and event["version"] == "2.0":
        return True

    return False


def format_response_to_aws(event: AwsEvent, response: HttpResponse) -> Dict[str, Any]:
    serverless_response: Dict[str, Any] = {"statusCode": int(response.status_code)}

    if "multiValueHeaders" in event:
        serverless_response["multiValueHeaders"] = response.headers._headers
    else:
        serverless_response["headers"] = {key: value for key, value in response.headers.items()}

    # If the request comes from ALB we need to add a status description
    is_elb = event.get("requestContext", {}).get("elb")
    if is_elb:
        serverless_response["statusDescription"] = str(response.status_code)

    content_type_header = response.headers.get("content-type", "text/plain")
    if not isinstance(content_type_header, str):
        content_type_header = content_type_header[0]
    mimetype, content_type_options = parse_header(content_type_header)
    body = str(response)

    if (mimetype.startswith("text/") or mimetype in TEXT_MIME_TYPES) and not response.headers.get(
        "Content-Encoding", ""
    ):

        serverless_response["body"] = body
        serverless_response["isBase64Encoded"] = False
    else:
        serverless_response["body"] = base64.b64encode(body.encode("utf8"))
        serverless_response["isBase64Encoded"] = True

    return serverless_response


def create_http_request_from_aws_event(event: AwsEvent, context: AwsContext) -> HttpRequest:
    is_http_api = is_http_api_lambda(event)

    if is_http_api:
        request = create_http_request_from_aws_http_api(event, context)
    else:
        request = create_http_request_from_aws_rest_api(event, context)

    return request


def create_http_request_from_aws_http_api(event: AwsEvent, context: AwsContext) -> HttpRequest:
    body = get_normalised_body_from_aws(event)
    request_context = event.get("requestContext", {})
    http_context = request_context.get("http", {})

    headers = get_normalised_headers_from_aws(event)
    headers["Cookie"] = "; ".join(event.get("cookies", []))
    headers["Content-Length"] = str(body.getbuffer().nbytes)

    request = HttpRequest(
        method=http_context["method"],
        path=http_context["path"],
        body=body,
        query_string=HttpQueryString(event.get("rawQueryString", "")),
        headers=HttpHeaders(headers),
    )
    request.path_parameters = event.get("pathParameters", {})

    request.attributes["aws_context"] = context
    request.attributes["aws_event"] = event

    return request


def create_http_request_from_aws_rest_api(event: AwsEvent, context: AwsContext) -> HttpRequest:
    body = get_normalised_body_from_aws(event)

    headers = get_normalised_headers_from_aws(event)
    headers["Content-Length"] = str(body.getbuffer().nbytes)

    raw_query_string = ""
    if event.get("multiValueQueryStringParameters"):
        for key, values in event.get("multiValueQueryStringParameters", {}).items():
            for value in values:
                raw_query_string += f"&{key}={quote_plus(value)}"

        raw_query_string = raw_query_string[1:]

    request = HttpRequest(
        method=event.get("httpMethod", "GET"),
        path=event.get("path", "/"),
        body=body,
        query_string=HttpQueryString(raw_query_string),
        headers=HttpHeaders(headers),
    )
    request.path_parameters = event.get("pathParameters", {})

    request.attributes["aws_context"] = context
    request.attributes["aws_event"] = event

    return request


def get_normalised_headers_from_aws(event: AwsEvent) -> Dict[str, str]:
    headers = event["headers"]
    request_context = event.get("requestContext", {})  # Set serverless related additional headers
    if request_context.get("requestId"):
        headers["x-serverless-request-id"] = request_context.get("requestId")
    if request_context.get("stage"):
        headers["x-serverless-stage"] = request_context.get("stage")
    if headers.get("x-amzn-trace-id"):
        headers["x-serverless-trace-id"] = headers["x-amzn-trace-id"]

    return headers


def get_normalised_body_from_aws(event: AwsEvent) -> BytesIO:
    body = event.get("body", "")
    if event.get("isBase64Encoded", False):
        body = base64.b64decode(body)

    if isinstance(body, str):
        body = body.encode("utf-8")

    return BytesIO(body)


__all__ = [
    "AwsEvent",
    "AwsContext",
    "AwsServerlessFunction",
    "create_http_request_from_aws_event",
]



class HttpError(Exception):
    status_code: int = 500
    http_message = "Internal Server Error"

    def __str__(self) -> str:
        return self.http_message

    def __repr__(self) -> str:
        return self.__class__.__name__ + f"({self.status_code}, {self.http_message})"


class NotFoundError(HttpError):
    status_code: int = 404
    http_message = "Not Found"


class BadRequestError(HttpError):
    status_code: int = 400
    http_message = "Bad Request"


__all__ = ["HttpError", "NotFoundError", "BadRequestError"]



from typing import Any, Callable, Dict, List, Optional, Union
import re

from .errors import (
    AdditionalPropertyError,
    MaximumPropertyError,
    MinimumPropertyError,
    PropertyNameError,
    PropertyValueError,
    RequiredPropertyError,
    ValidationError,
    MissingDependencyError,
)


def validate_object_properties(
    obj: dict,
    properties: Dict[str, Callable] = None,
    additional_properties: Union[bool, Callable] = True,
    pattern_properties: Dict[str, Callable] = None,
) -> dict:
    properties = properties if properties is not None else {}
    for key, value in obj.items():
        if key in properties:
            try:
                obj[key] = properties[key](value)
                continue
            except PropertyValueError as error:
                raise PropertyValueError(
                    sub_code=error.code,
                    property_name=key + "." + error.context["property_name"],
                    validation_error=error.context["validation_error"],
                )
            except ValidationError as error:
                raise PropertyValueError(property_name=key, validation_error=str(error), sub_code=error.code)
            except ValueError as error:
                raise PropertyValueError(property_name=key, validation_error=str(error))

        if not additional_properties and not pattern_properties:
            raise AdditionalPropertyError(property_name=key)

        if pattern_properties:
            property_validator: Optional[Callable] = None
            for property_pattern, tmp_validator in pattern_properties.items():
                if re.search(property_pattern, key):
                    property_validator = tmp_validator
                    break

            if property_validator:
                obj[key] = property_validator(value)
                continue

            if additional_properties is False:
                raise AdditionalPropertyError(property_name=key)

        if additional_properties is False:
            raise AdditionalPropertyError(property_name=key)

        if additional_properties is True:
            continue

        try:
            obj[key] = additional_properties(value)  # type: ignore
        except PropertyValueError as error:
            raise PropertyValueError(
                property_name=key + "." + error.context["property_name"],
                validation_error=error.context["validation_error"],
            )
        except ValidationError as error:
            raise PropertyValueError(
                property_name=key, validation_error=str(error), sub_code=error.code,
            )
        except ValueError as error:
            raise PropertyValueError(property_name=key, validation_error=str(error))

    return obj


def validate_object_property_names(obj: dict, property_names: Callable) -> dict:
    for name in obj.keys():
        try:
            property_names(name)
        except ValidationError as error:
            raise PropertyNameError(sub_code=error.code, property_name=name, validation_error=str(error))
        except ValueError as error:
            raise PropertyNameError(sub_code="error", property_name=name, validation_error=str(error))

    return obj


def validate_object_required_properties(obj: dict, required_properties: List[str]) -> dict:
    given_properties = obj.keys()
    missing_properties = [key for key in required_properties if key not in given_properties]

    if missing_properties:
        raise RequiredPropertyError(property_name=missing_properties[0])

    return obj


def validate_object_minimum_properties(obj: dict, expected_minimum: int) -> dict:
    if len(obj) < expected_minimum:
        raise MinimumPropertyError(expected_minimum=expected_minimum)
    return obj


def validate_object_maximum_properties(obj: dict, expected_maximum: int) -> dict:
    if len(obj) > expected_maximum:
        raise MaximumPropertyError(expected_maximum=expected_maximum)

    return obj


def validate_object_dependencies(obj: dict, dependencies: Dict[str, List[str]]) -> dict:
    for field_name, field_dependencies in dependencies.items():
        if field_name not in obj:
            continue
        if not all(k in obj for k in field_dependencies):
            raise MissingDependencyError(property=field_name, dependencies=field_dependencies)

    return obj


__all__ = [
    "validate_object_dependencies",
    "validate_object_maximum_properties",
    "validate_object_minimum_properties",
    "validate_object_properties",
    "validate_object_property_names",
    "validate_object_required_properties",
]
